/**
 * Introduction
 *
 * @title ROV - Easy applications from C
 *
 * @copyright Jason Tribbeck / 7th ARM 1994-2001
 *
 * @authors Jason Tribbeck
 *
 * @file vti+.c
 */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>

#include "kernel.h"

#ifndef __GNUC__
#pragma -s1
#endif

#include "main.h"
#include "canvas.h"
#include "icons.h"
#include "menu.h"
#include "msg.h"
#include "extras.h"

#include "vtiinet.h"

extern canvas VTI_FOREC,
  VTI_BACKC;

extern int VTI_CWOX, VTI_CWOY, VTI_CWWIDTH, VTI_CWHEIGHT;

static int VTI_AWWIDTH, _task;

static icon _i[64][10], bi, mi, ti;
static menu_make _m[64];
static list_addop _a[64];
static char help[64][4];
static int _h[96], _ih = 0;

/* static icon first_writeable = 0; */
static int ic = 0, Y = 0, height = 0;

static char null_text[4] = "";
static char groupval2[4] = "r1";
static char indval[4] = "r2";
/* static char groupval[4] = "r2"; */
static char actionval[8] = "r5,3";
static char upval[16]	 = "sup,pup";
static char downval[16]	 = "sdown,pdown";
static char optval[16] = "soptoff,opton";
static char radval[20] = "sradiooff,radioon";
static char valid[24] = "kartn;pptr_write,4,6";
static char menuval[24] = "sgright;pptr_menu,2,2";
static char numval[32] = "a0-9;kartn;pptr_write,4,6";
static char ipval[32] = "a0-9.;kartn;pptr_write,4,6";
static char alpval[32] = "aA-Za-z;kartn;pptr_write,4,6";
static char alnumval[32] = "aA-Za-z0-9;kartn;pptr_write,4,6";
static char hexval[32] = "aA-Fa-f0-9;kartn;pptr_write,4,6";
static char ngroupval[44] = "aA-Za-z0-9.*@:\\-\\\\/;kartn;pptr_write,4,6";
static _opts cancel, write;
static int _short = 0;

int VTiInternet_ConfigureWindows(char * title, int h);

static void open(canvas * c, canvas_openstr * o)
{
  canvas_openstr o2;

  canvas_getcanvasstate(VTI_FOREC, & o2);
  memcpy(& o2, o, 16);
  o2.x0 += VTI_CWOX;
  o2.y0 += VTI_CWOY;
  o2.x1 = o2.x0 + VTI_CWWIDTH;
  o2.y1 = o2.y0 + VTI_CWHEIGHT;
  o2.behind = o -> behind;
  canvas_open(VTI_FOREC, & o2);
  o -> behind = canvas_handle(VTI_FOREC);
  canvas_open(VTI_BACKC, o);
}

/* static void ukh(int i, message_block * b)
   {
   if(b ->
   } */

static void _wr(canvas c, icon i, canvas_mousestr * k)
{
  message_block m;
  write();
  m.data.words[0] = _task;
  m.my_ref = m.your_ref = 0;
  if((unsigned int)_task < 32) {
    m.action = 0x4a435;
  } else {
    m.action = 0x4a424;
  }
  m.size = 24;
  main_send(& m, 17, 0, 0);
  if(k -> b == 4) cancel();
}

void VTi_Help(void)
{
  char l[8], * b = 0, * c;
  int i;

  c = poll_block.data.chars;
  if(poll_block.data.words[0] == canvas_handle(VTI_FOREC)) {
    for(i = 0; i < _ih; i ++) {
      if(poll_block.data.words[1] == _h[i]) {
	sprintf(l, "B%s", help[i]);
	b = msg_lookup(l);
	goto flip;
      }
    }
    b = msg_lookuprom("BH0");
    goto flip;
  }
  if(poll_block.data.words[0] == canvas_handle(VTI_BACKC)) {
    if(/* (poll_block.data.words[1] > 3) || */(poll_block.data.words[1] < 0)) {
      b = msg_lookuprom("BH0");
    } else {
      sprintf(l, "BH0%d", poll_block.data.words[1]);
      b = msg_lookuprom(l);
    }
  }
  if(!b) return;
 flip:
  while(* b > 31) * c ++ = * b ++ ;
  * c = 0;
  poll_block.size = (24 + strlen(poll_block.data.chars)) & 0xfffffffc;
  poll_block.your_ref = poll_block.my_ref;
  poll_block.action = 0x4a43b;
  main_send(& poll_block, 17, poll_block.sender, 0);
}

static void VTi_Help2(void)
{
  char l[8];
  int i;

  for(i = 0; i < _ih; i ++) {
    if(poll_block.data.words[4] == _h[i]) {
      sprintf(l, "I%s", help[i]);
      return_help(msg_lookuprom(l));
      return;
    }
  }
  return_help(msg_lookuprom("IH0"));
}

static void VTi_Help1(void)
{
  char l[8];
  /* int i; */

  if((poll_block.data.words[4] > 2) || (poll_block.data.words[4] < 0)) {
    return_help(msg_lookuprom("IH0"));
  } else {
    sprintf(l, "IH0%d", poll_block.data.words[4]);
    return_help(msg_lookuprom(l));
  }
}

static int hh;

static void cczz(void)
{
}

void VTi_MakeRedraw(void)
{
  canvas_openstr o;

  canvas_getcanvasstate(VTI_FOREC, & o);
  canvas_removeallocated(VTI_FOREC);
  VTI_FOREC = canvas_create(o.x1 - o.x0, VTI_CWHEIGHT, "");
  canvas_setflags(VTI_FOREC, 0xff000000, 0x80000000);
  if(hh > VTI_CWHEIGHT) {
    canvas_setfullsize(VTI_FOREC, 0, 0, o.x1 - o.x0, hh);
    canvas_setflags(VTI_FOREC, 0xff000000, 0x90000000);
  }
  canvas_addhandler(VTI_FOREC, CANVAS_REDRAWHANDLER, cczz);
  canvas_make(VTI_FOREC);
  canvas_open(VTI_FOREC, & o);
}

void VTi_OpenConfigure(char * title, int h, _opts wr, _opts rd, _opts cn, int task)
{
  canvas_openstr o;
  message_block b;

  write = wr;
  cancel = cn;

  if(VTiInternet_ConfigureWindows(title, h) == 0) {
    simple_warning("Unable to create configuration windows", 1);
    exit(0);
  }

  if(h < 480) h = 480;
  hh = h;
  Y = h - 12;

  height = h;

  canvas_addhandler(VTI_BACKC, CANVAS_OPENHANDLER, open);

  VTI_AWWIDTH = VTI_CWWIDTH;
  if(h > 480) VTI_AWWIDTH -= 38;

  /* d.indirecttext.text = null_text;
     d.indirecttext.valid = groupval;
     d.indirecttext.length = 12;
     icon_createforgetful(VTI_FOREC, 0, 0, VTI_AWWIDTH, h, 0x1700011d, & d); */

  canvas_opencentre(VTI_BACKC);
  canvas_getcanvasstate(VTI_BACKC, & o);

  o.x0 += VTI_CWOX;
  o.y0 += VTI_CWOY;
  o.scy = height;
  o.x1 = o.x0 + VTI_CWWIDTH;
  o.y1 = o.y0 + VTI_CWHEIGHT;
  o.behind = -1;
  canvas_open(VTI_FOREC, & o);

  if((unsigned int)task < 32) {
    b.action = 0x4a43f;
    b.size = 24;
  } else {
    b.action = 0x4a423;
    b.size = 28;
    b.data.words[1] = 0;
  }
  b.my_ref = b.your_ref = 0;
  b.data.words[0] = _task = task;

  main_send(& b, 17, 0, 0);

  bi = icon_useexisting(VTI_BACKC, 0);
  icon_addhandler(bi, ICON_CLICKHANDLER, _wr);
  ti = icon_useexisting(VTI_BACKC, 1);
  icon_addhandler(ti, ICON_CLICKHANDLER, rd);
  mi = icon_useexisting(VTI_BACKC, 2);
  icon_addhandler(mi, ICON_CLICKHANDLER, cn);
  canvas_addhandler(VTI_BACKC, CANVAS_HELPHANDLER, VTi_Help1);
  canvas_addhandler(VTI_FOREC, CANVAS_HELPHANDLER, VTi_Help2);
  canvas_addhandler(VTI_BACKC, CANVAS_CLOSEHANDLER, cn);
}

static void ick_h(canvas c, icon i, canvas_keystr * k)
{
  if(k -> chcode > 0x100) {
    main_processkey(k -> chcode);
  } else {
    if((k -> chcode == 13) || (k -> chcode == 10)) {
      icon_select(bi, 1);
      write();
      icon_select(bi, 0);
      cancel();
    }
    if(k -> chcode == 0x1b) {
      cancel();
    }
  }
}

static void cm_m(canvas c, icon icn)
{
  int i, j;
  menu m;

  for(i = 0; i < ic; i ++) {
    if(_i[i][6]) j = 6; else j = 0;
    if(icn == _i[i][j]) {
      m = _m[i](_i[i][0]);
      if(m)
	menu_displayicon(m, _i[i][1]);
      return;
    }
  }

  for(i = 0; i < ic; i ++) {
    if(icn == _i[i][1]) {
      m = _m[i](_i[i][0]);
      if(m)
	menu_displayicon(m, _i[i][1]);
      return;
    }
  }
}

icon VTi_CreateMenu(char * txt, char * box, menu_make m, char * hlp)
{
  icon i, i2;
  icon_data d;

  memcpy(d.text, txt, 12);
  icon_createforgetful(VTI_FOREC, 0, Y - 44, 192, Y, 0x17000211, & d);
  d.indirecttext.text = box;
  d.indirecttext.valid = indval;
  d.indirecttext.length = 12;
  i = icon_create(VTI_FOREC, 192, Y - 44, VTI_AWWIDTH - 56, Y, 0x1700313d, & d);
  d.indirecttext.text = null_text;
  d.indirecttext.valid = menuval;
  i2 = icon_create(VTI_FOREC, VTI_AWWIDTH - 52, Y - 44, VTI_AWWIDTH - 8, Y, 0x1700311b, & d);

  icon_addhandler(i, ICON_MENUHANDLER, cm_m);
  icon_addhandler(i2, ICON_MENUHANDLER, cm_m);
  icon_addhandler(i, ICON_CLICKHANDLER, cm_m);
  icon_addhandler(i2, ICON_CLICKHANDLER, cm_m);
  _h[_ih] = icon_handle(i);
  memcpy(help[_ih ++ ], hlp, 4);
  _h[_ih] = icon_handle(i2);
  memcpy(help[_ih ++ ], hlp, 4);

  Y -= 52;

  _m[ic] = m;
  _i[ic][0] = i;
  _i[ic ++ ][1] = i2;

  return i;
}

icon VTi_CreateLongMenu(char * txt, char * box, menu_make m, char * hlp)
{
  icon i, i2;
  icon_data d;

  memcpy(d.text, txt, 12);
  icon_createforgetful(VTI_FOREC, 0, Y - 44, 640, Y, 0x17000011, & d);
  Y -= 48;
  d.indirecttext.text = box;
  d.indirecttext.valid = indval;
  d.indirecttext.length = 12;
  i = icon_create(VTI_FOREC, 8, Y - 44, VTI_AWWIDTH - 56, Y, 0x1700313d, & d);
  d.indirecttext.text = null_text;
  d.indirecttext.valid = menuval;
  i2 = icon_create(VTI_FOREC, VTI_AWWIDTH - 52, Y - 44, VTI_AWWIDTH - 8, Y, 0x1700311b, & d);

  icon_addhandler(i, ICON_MENUHANDLER, cm_m);
  icon_addhandler(i2, ICON_MENUHANDLER, cm_m);
  icon_addhandler(i, ICON_CLICKHANDLER, cm_m);
  icon_addhandler(i2, ICON_CLICKHANDLER, cm_m);
  _h[_ih] = icon_handle(i);
  memcpy(help[_ih ++ ], hlp, 4);
  _h[_ih] = icon_handle(i2);
  memcpy(help[_ih ++ ], hlp, 4);

  Y -= 52;

  _m[ic] = m;
  _i[ic][0] = i;
  _i[ic ++ ][1] = i2;

  return i;
}

void VTi_CreateHBar(void)
{
  icon_data d;

  d.indirecttext.text = null_text;
  d.indirecttext.valid = groupval2;
  d.indirecttext.length = 12;
  icon_createforgetful(VTI_FOREC, - 8, Y - 8, VTI_AWWIDTH + 8, Y, 0x1700011d, & d);
  Y -= 16;
}

void VTi_CreateComment(char * txt)
{
  icon_data d;

  d.indirecttext.text = txt;
  d.indirecttext.valid = (char *) - 1;
  d.indirecttext.length = 12;
  icon_createforgetful(VTI_FOREC, 8, Y - 44, VTI_AWWIDTH - 8, Y, 0x17000111, & d);
  Y -= 44;
}

icon VTi_CreateOption(char * txt, int ESG, char * hlp)
{
  icon_data d;
  icon i;

  d.indirecttext.text = txt;
  if(ESG)
    d.indirecttext.valid = radval;
  else
    d.indirecttext.valid = optval;
  d.indirecttext.length = 12;
  i = icon_create(VTI_FOREC, 8, Y - 44, VTI_AWWIDTH - 8, Y, 0x17003113 | (ESG << 16), & d);
  memcpy(help[_ih], hlp, 4);
  _h[_ih ++ ] = icon_handle(i);
  Y -= 52;
  return i;
}

icon VTi_CreateShortOption(char * txt, int ESG, char * hlp)
{
  icon_data d;
  icon i;

  d.indirecttext.text = txt;
  if(ESG)
    d.indirecttext.valid = radval;
  else
    d.indirecttext.valid = optval;
  d.indirecttext.length = 12;
  if(_short) {
    i = icon_create(VTI_FOREC, 8, Y - 44, (VTI_AWWIDTH >> 1) - 8, Y, 0x17003113 | (ESG << 16), & d);
  } else {
    i = icon_create(VTI_FOREC, (VTI_AWWIDTH >> 1) + 8, Y - 44, VTI_AWWIDTH - 8, Y, 0x17003113 | (ESG << 16), & d);
  }
  memcpy(help[_ih], hlp, 4);
  _h[_ih ++ ] = icon_handle(i);
  if(_short) Y -= 52;
  _short ^= 1;
  return i;
}

void VTi_ReleaseShort(void)
{
  if(_short) Y -= 52;
  _short = 0;
}

void VTi_FadeMenu(icon icn, int fl)
{
  int i;

  for(i = 0; i < ic; i ++) {
    if(icn == _i[i][0]) {
      icon_fade(_i[i][0], fl);
      icon_fade(_i[i][1], fl);
    }
  }
}

icon VTi_CreateUnits(char * txt, char * box, char * units, int len, int type, char * hlp)
{
  icon_data d;
  icon i;

  memcpy(d.text, txt, 12);
  icon_createforgetful(VTI_FOREC, 0, Y - 44, 192, Y, 0x17000211, & d);
  memcpy(d.text, units, 12);
  icon_createforgetful(VTI_FOREC, VTI_AWWIDTH - 96, Y - 44, VTI_AWWIDTH, Y, 0x17000011, & d);
  d.indirecttext.text = box;
  d.indirecttext.length = len;

  switch(type) {
  case VTI_ALL :
    d.indirecttext.valid = valid;
    break;
  case VTI_DIGITS :
    d.indirecttext.valid = numval;
    break;
  case VTI_ALPHA :
    d.indirecttext.valid = alpval;
    break;
  case VTI_ALNUM :
    d.indirecttext.valid = alnumval;
    break;
  case VTI_HEX :
    d.indirecttext.valid = hexval;
    break;
  }

  i = icon_create(VTI_FOREC, 192, Y - 44, VTI_AWWIDTH - 96, Y, 0x0700f13d, & d);
  icon_addhandler(i, ICON_KEYHANDLER, ick_h);
  memcpy(help[_ih], hlp, 4);
  _h[_ih ++ ] = icon_handle(i);
  /* if(!first_writeable) icon_setcaret(first_writeable = i, 0, 0, 0); */

  Y -= 52;

  return i;
}

icon VTi_CreateInformation(char * txt, char * box, char * hlp)
{
  icon i;
  icon_data d;

  memcpy(d.text, txt, 12);
  icon_createforgetful(VTI_FOREC, 0, Y - 44, 192, Y, 0x17000211, & d);
  d.indirecttext.text = box;
  d.indirecttext.valid = indval;
  d.indirecttext.length = 12;
  i = icon_create(VTI_FOREC, 192, Y - 44, VTI_AWWIDTH - 8, Y, 0x1700313d, & d);
  memcpy(help[_ih], hlp, 4);
  _h[_ih ++ ] = icon_handle(i);

  Y -= 52;
  return i;
}

icon VTi_CreateLargeInformation(char * txt, char * box, char * hlp)
{
  icon i;
  icon_data d;

  memcpy(d.text, txt, 12);
  icon_createforgetful(VTI_FOREC, 0, Y - 44, VTI_AWWIDTH, Y, 0x17000011, & d);
  d.indirecttext.text = box;
  d.indirecttext.valid = indval;
  d.indirecttext.length = 12;

  Y -= 48;
  i = icon_create(VTI_FOREC, 8, Y - 44, VTI_AWWIDTH - 8, Y, 0x1700313d, & d);
  memcpy(help[_ih], hlp, 4);
  _h[_ih ++ ] = icon_handle(i);

  Y -= 52;
  return i;
}

static void cu_m(canvas c, icon icn)
{
  int i;

  for(i = 0; i < ic; i ++) {
    if(icn == _i[i][2]) {
      if(_a[i](LISTOP_UP)) {
	icon_select(_i[i][2], 1);
	VTi_VSync();
	icon_select(_i[i][2], 0);
      }
      return;
    }
  }
}

static void cd_m(canvas c, icon icn)
{
  int i;

  for(i = 0; i < ic; i ++) {
    if(icn == _i[i][3]) {
      if(_a[i](LISTOP_DOWN)) {
	icon_select(_i[i][3], 1);
	VTi_VSync();
	icon_select(_i[i][3], 0);
      }
      return;
    }
  }
}

static void ca_m(canvas c, icon icn)
{
  int i;

  for(i = 0; i < ic; i ++) {
    if(icn == _i[i][4]) {
      _a[i](LISTOP_ADD);
      return;
    }
  }
}

static void cr_m(canvas c, icon icn)
{
  int i;

  for(i = 0; i < ic; i ++) {
    if(icn == _i[i][5]) {
      _a[i](LISTOP_REMOVE);
      return;
    }
  }
}

static void ck_m(canvas c, icon i, canvas_keystr * k)
{
  int n;

  if(k -> chcode == 0x1b) {
    cancel();
  }

  if((k -> chcode == 13) || (k -> chcode == 10)) {
    for(n = 0; n < ic; n ++) {
      if(i == _i[n][0]) {
	if(_a[n](LISTOP_ADD)) {
	  icon_select(_i[n][4], 1);
	  VTi_VSync();
	  icon_select(_i[n][4], 0);
	}
	return;
      }
    }
  }

  if((k -> chcode > 0x100) || (k -> chcode == 0x7f) || (k -> chcode < 32)) {
    main_processkey(k -> chcode);
    return;
  }
}

static void ck2_m(canvas c, icon i, canvas_keystr * k)
{
  if(k -> chcode == 0x1b) {
    cancel();
  }

  if((k -> chcode > 0x100) || (k -> chcode == 0x7f) || (k -> chcode < 32)) {
    main_processkey(k -> chcode);
    return;
  }
}

static void Setup(icon_data * d, int t, char * s, int l)
{
  d -> indirecttext.text = s;
  switch(t) {
  case VTI_DIGITS :
    d -> indirecttext.valid = numval;
    break;
  case VTI_ALPHA :
    d -> indirecttext.valid = alpval;
    break;
  case VTI_ALNUM :
    d -> indirecttext.valid = alnumval;
    break;
  case VTI_HEX :
    d -> indirecttext.valid = hexval;
    break;
  case VTI_GROUP :
    d -> indirecttext.valid = ngroupval;
    break;
  default :
    d -> indirecttext.valid = valid;
    break;
  }
  d -> indirecttext.length = l;
}

icon VTi_CreateExtendibleList(char * txt, char * box, menu_make m, list_addop a, int size, int type, char * hlp)
{
  icon i, i2, i3, i4, i5, i6;
  icon_data d;

  memcpy(d.text, txt, 12);
  icon_createforgetful(VTI_FOREC, 0, Y - 44, 192, Y, 0x17000211, & d);
  Setup(& d, type, box, size);
  i = icon_create(VTI_FOREC, 192 + 40, Y - 44, VTI_AWWIDTH - 56 - 44, Y, 0x0700f13d, & d);
  d.indirecttext.text = null_text;
  d.indirecttext.valid = menuval;
  i2 = icon_create(VTI_FOREC, VTI_AWWIDTH - 52, Y - 44, VTI_AWWIDTH - 8, Y, 0x1700311b, & d);
  d.indirecttext.valid = upval;
  i3 = icon_create(VTI_FOREC, 192, Y - 44, 192 + 40, Y, 0x1700211b, & d);
  d.indirecttext.valid = downval;
  i4 = icon_create(VTI_FOREC, VTI_AWWIDTH - 52 - 40, Y - 44, VTI_AWWIDTH - 56, Y, 0x1700211b, & d);
  d.indirecttext.valid = actionval;
  d.indirecttext.text = msg_lookuprom("Add");

  Y -= 52;

  i5 = icon_create(VTI_FOREC, VTI_AWWIDTH - 256 - 16, Y - 44, VTI_AWWIDTH - 160 - 16, Y, 0x1700313d, & d);
  d.indirecttext.text = msg_lookuprom("Remove");
  i6 = icon_create(VTI_FOREC, VTI_AWWIDTH - 160 - 8, Y - 44, VTI_AWWIDTH - 8, Y, 0x1700313d, & d);

  icon_addhandler(i, ICON_MENUHANDLER, cm_m);
  icon_addhandler(i, ICON_KEYHANDLER, ck_m);
  icon_addhandler(i2, ICON_MENUHANDLER, cm_m);
  icon_addhandler(i2, ICON_CLICKHANDLER, cm_m);
  icon_addhandler(i3, ICON_CLICKHANDLER, cu_m);
  icon_addhandler(i4, ICON_CLICKHANDLER, cd_m);
  icon_addhandler(i5, ICON_CLICKHANDLER, ca_m);
  icon_addhandler(i6, ICON_CLICKHANDLER, cr_m);
  _h[_ih] = icon_handle(i);
  memcpy(help[_ih ++ ], hlp, 4);
  _h[_ih] = icon_handle(i2);
  memcpy(help[_ih ++ ], hlp, 4);
  _h[_ih] = icon_handle(i3);
  memcpy(help[_ih ++ ], hlp, 4);
  _h[_ih] = icon_handle(i4);
  memcpy(help[_ih ++ ], hlp, 4);
  _h[_ih] = icon_handle(i5);
  memcpy(help[_ih ++ ], hlp, 4);
  _h[_ih] = icon_handle(i6);
  memcpy(help[_ih ++ ], hlp, 4);

  Y -= 52;

  _m[ic] = m;
  _a[ic] = a;
  _i[ic][0] = i;
  _i[ic][1] = i2;
  _i[ic][2] = i3;
  _i[ic][3] = i4;
  _i[ic][4] = i5;
  _i[ic ++ ][5] = i6;

  return i;
}

void VTi_FadeExtendibleList(icon icn, int fl)
{
  int i, j;

  for(i = 0; i < ic; i ++) {
    if(icn == _i[i][0]) {
      if(fl == LIST_FULLFADE) {
	icon_fade(_i[i][0], fl);
	icon_fade(_i[i][4], fl);
	if(_i[i][6])
	  for(j = 6; j < 10; j ++)
	    if(_i[i][j]) icon_fade(_i[i][j], fl);
      } else {
	icon_fade(_i[i][0], 0);
	icon_fade(_i[i][4], 0);
	if(_i[i][6])
	  for(j = 6; j < 10; j ++)
	    if(_i[i][j]) icon_fade(_i[i][j], 0);
      }
      icon_fade(_i[i][1], fl);
      icon_fade(_i[i][2], fl);
      icon_fade(_i[i][3], fl);
      icon_fade(_i[i][5], fl);
    }
  }
}

icon VTi_CreateLargeExtendibleList(char * txt, char * box, menu_make m, list_addop a, int size, int type, char * hlp)
{
  icon i, i2, i3, i4, i5, i6;
  icon_data d;

  d.indirecttext.text = txt;
  d.indirecttext.valid = (char *) - 1;
  d.indirecttext.length = 12;

  icon_createforgetful(VTI_FOREC, 0, Y - 44, VTI_AWWIDTH - 8, Y, 0x17000111, & d);
  Y -= 48;
  Setup(& d, type, box, size);
  i = icon_create(VTI_FOREC, 8 + 40, Y - 44, VTI_AWWIDTH - 56 - 44, Y, 0x0700f13d, & d);
  d.indirecttext.text = null_text;
  d.indirecttext.valid = menuval;
  i2 = icon_create(VTI_FOREC, VTI_AWWIDTH - 52, Y - 44, VTI_AWWIDTH - 8, Y, 0x1700311b, & d);
  d.indirecttext.valid = upval;
  i3 = icon_create(VTI_FOREC, 8, Y - 44, 8 + 40, Y, 0x1700211b, & d);
  d.indirecttext.valid = downval;
  i4 = icon_create(VTI_FOREC, VTI_AWWIDTH - 52 - 40, Y - 44, VTI_AWWIDTH - 56, Y, 0x1700211b, & d);
  d.indirecttext.valid = actionval;
  d.indirecttext.text = msg_lookuprom("Add");

  Y -= 52;

  i5 = icon_create(VTI_FOREC, VTI_AWWIDTH - 256 - 16, Y - 44, VTI_AWWIDTH - 160 - 16, Y, 0x1700313d, & d);
  d.indirecttext.text = msg_lookuprom("Remove");
  i6 = icon_create(VTI_FOREC, VTI_AWWIDTH - 160 - 8, Y - 44, VTI_AWWIDTH - 8, Y, 0x1700313d, & d);

  icon_addhandler(i, ICON_MENUHANDLER, cm_m);
  icon_addhandler(i, ICON_KEYHANDLER, ck_m);
  icon_addhandler(i2, ICON_MENUHANDLER, cm_m);
  icon_addhandler(i2, ICON_CLICKHANDLER, cm_m);
  icon_addhandler(i3, ICON_CLICKHANDLER, cu_m);
  icon_addhandler(i4, ICON_CLICKHANDLER, cd_m);
  icon_addhandler(i5, ICON_CLICKHANDLER, ca_m);
  icon_addhandler(i6, ICON_CLICKHANDLER, cr_m);
  _h[_ih] = icon_handle(i);
  memcpy(help[_ih ++ ], hlp, 4);
  _h[_ih] = icon_handle(i2);
  memcpy(help[_ih ++ ], hlp, 4);
  _h[_ih] = icon_handle(i3);
  memcpy(help[_ih ++ ], hlp, 4);
  _h[_ih] = icon_handle(i4);
  memcpy(help[_ih ++ ], hlp, 4);
  _h[_ih] = icon_handle(i5);
  memcpy(help[_ih ++ ], hlp, 4);
  _h[_ih] = icon_handle(i6);
  memcpy(help[_ih ++ ], hlp, 4);

  Y -= 52;

  _m[ic] = m;
  _a[ic] = a;
  _i[ic][0] = i;
  _i[ic][1] = i2;
  _i[ic][2] = i3;
  _i[ic][3] = i4;
  _i[ic][4] = i5;
  _i[ic ++ ][5] = i6;

  return i;
}

icon VTi_CreateAction(char * string, char * hlp)
{
  icon i;
  icon_data d;

  d.indirecttext.length = 32;
  d.indirecttext.valid = actionval;
  d.indirecttext.text = string;
  i = icon_create(VTI_FOREC, 8, Y - 44, VTI_AWWIDTH - 8, Y, 0x1700313d, & d);

  _h[_ih] = icon_handle(i);
  memcpy(help[_ih ++ ], hlp, 4);

  Y -= 52;
  return i;
}

static char hlp[4], opencount = 0;

icon VTi_CreateOpenExtendibleList(char * txt, char * box, menu_make m, list_addop a, int size, int type, char * hlpq)
{
  icon i, i2, i3, i4;
  icon_data d;

  memcpy(hlp, hlpq, 4);
  memcpy(d.text, txt, 12);
  icon_createforgetful(VTI_FOREC, 0, Y - 44, 192, Y, 0x17000211, & d);
  Setup(& d, type, box, size);
  i = icon_create(VTI_FOREC, 192 + 40, Y - 44, VTI_AWWIDTH - 56 - 44, Y, 0x0700f13d, & d);
  d.indirecttext.text = null_text;
  d.indirecttext.valid = menuval;
  i2 = icon_create(VTI_FOREC, VTI_AWWIDTH - 52, Y - 44, VTI_AWWIDTH - 8, Y, 0x1700311b, & d);
  d.indirecttext.valid = upval;
  i3 = icon_create(VTI_FOREC, 192, Y - 44, 192 + 40, Y, 0x1700211b, & d);
  d.indirecttext.valid = downval;
  i4 = icon_create(VTI_FOREC, VTI_AWWIDTH - 52 - 40, Y - 44, VTI_AWWIDTH - 56, Y, 0x1700211b, & d);

  Y -= 52;

  icon_addhandler(i, ICON_MENUHANDLER, cm_m);
  icon_addhandler(i, ICON_KEYHANDLER, ck2_m);
  icon_addhandler(i2, ICON_MENUHANDLER, cm_m);
  icon_addhandler(i2, ICON_CLICKHANDLER, cm_m);
  icon_addhandler(i3, ICON_CLICKHANDLER, cu_m);
  icon_addhandler(i4, ICON_CLICKHANDLER, cd_m);
  _h[_ih] = icon_handle(i);
  memcpy(help[_ih ++ ], hlp, 4);
  _h[_ih] = icon_handle(i2);
  memcpy(help[_ih ++ ], hlp, 4);
  _h[_ih] = icon_handle(i3);
  memcpy(help[_ih ++ ], hlp, 4);
  _h[_ih] = icon_handle(i4);
  memcpy(help[_ih ++ ], hlp, 4);

  _m[ic] = m;
  _a[ic] = a;
  _i[ic][0] = i;
  _i[ic][1] = i2;
  _i[ic][2] = i3;
  _i[ic][3] = i4;
  _i[ic][6] = i;

  opencount = 7;

  return i;
}

icon VTi_CreateOpenLargeExtendibleList(char * txt, char * box, menu_make m, list_addop a, int size, int type, char * hlpq)
{
  icon i, i2, i3, i4;
  icon_data d;

  d.indirecttext.text = txt;
  d.indirecttext.valid = (char *) - 1;
  d.indirecttext.length = 12;

  icon_createforgetful(VTI_FOREC, 0, Y - 44, VTI_AWWIDTH - 8, Y, 0x17000111, & d);
  Y -= 48;

  Setup(& d, type, box, size);
  i = icon_create(VTI_FOREC, 8 + 40, Y - 44, VTI_AWWIDTH - 56 - 44, Y, 0x0700f13d, & d);
  d.indirecttext.text = null_text;
  d.indirecttext.valid = menuval;
  i2 = icon_create(VTI_FOREC, VTI_AWWIDTH - 52, Y - 44, VTI_AWWIDTH - 8, Y, 0x1700311b, & d);
  d.indirecttext.valid = upval;
  i3 = icon_create(VTI_FOREC, 8, Y - 44, 8 + 40, Y, 0x1700211b, & d);
  d.indirecttext.valid = downval;
  i4 = icon_create(VTI_FOREC, VTI_AWWIDTH - 52 - 40, Y - 44, VTI_AWWIDTH - 56, Y, 0x1700211b, & d);

  Y -= 52;

  icon_addhandler(i, ICON_MENUHANDLER, cm_m);
  icon_addhandler(i, ICON_KEYHANDLER, ck2_m);
  icon_addhandler(i2, ICON_MENUHANDLER, cm_m);
  icon_addhandler(i2, ICON_CLICKHANDLER, cm_m);
  icon_addhandler(i3, ICON_CLICKHANDLER, cu_m);
  icon_addhandler(i4, ICON_CLICKHANDLER, cd_m);
  _h[_ih] = icon_handle(i);
  memcpy(help[_ih ++ ], hlp, 4);
  _h[_ih] = icon_handle(i2);
  memcpy(help[_ih ++ ], hlp, 4);
  _h[_ih] = icon_handle(i3);
  memcpy(help[_ih ++ ], hlp, 4);
  _h[_ih] = icon_handle(i4);
  memcpy(help[_ih ++ ], hlp, 4);

  _m[ic] = m;
  _a[ic] = a;
  _i[ic][0] = i;
  _i[ic][1] = i2;
  _i[ic][2] = i3;
  _i[ic][3] = i4;
  _i[ic][6] = i;

  opencount = 7;

  return i;
}

icon VTi_AddToExtendibleList(char * txt, char * box, int size, int type)
{
  icon i;
  icon_data d;

  memcpy(d.text, txt, 12);
  icon_createforgetful(VTI_FOREC, 0, Y - 44, 192, Y, 0x17000211, & d);
  Setup(& d, type, box, size);
  i = icon_create(VTI_FOREC, 192, Y - 44, VTI_AWWIDTH - 56, Y, 0x0700f13d, & d);

  Y -= 52;

  icon_addhandler(i, ICON_MENUHANDLER, cm_m);
  icon_addhandler(i, ICON_KEYHANDLER, ck2_m);
  _h[_ih] = icon_handle(i);
  memcpy(help[_ih ++ ], hlp, 4);

  _i[ic][0] = i;
  _i[ic][(int)opencount ++] = i;

  return i;
}

void VTi_CloseExtendibleList(void)
{
  icon i5, i6;
  icon_data d;

  d.indirecttext.valid = actionval;
  d.indirecttext.text = msg_lookuprom("Add");
  i5 = icon_create(VTI_FOREC, VTI_AWWIDTH - 256 - 16, Y - 44, VTI_AWWIDTH - 160 - 16, Y, 0x1700313d, & d);
  d.indirecttext.text = msg_lookuprom("Remove");
  i6 = icon_create(VTI_FOREC, VTI_AWWIDTH - 160 - 8, Y - 44, VTI_AWWIDTH - 8, Y, 0x1700313d, & d);

  Y -= 52;

  _h[_ih] = icon_handle(i5);
  memcpy(help[_ih ++ ], hlp, 4);
  _h[_ih] = icon_handle(i6);
  memcpy(help[_ih ++ ], hlp, 4);

  icon_addhandler(i5, ICON_CLICKHANDLER, ca_m);
  icon_addhandler(i6, ICON_CLICKHANDLER, cr_m);
  icon_addhandler(_i[ic][0], ICON_KEYHANDLER, ck_m);
  _i[ic][4] = i5;
  _i[ic ++ ][5] = i6;
}

icon VTi_CreateOpenLargeInformation(char * txt, char * box)
{
  icon i;
  icon_data d;

  d.indirecttext.text = txt;
  d.indirecttext.valid = (char *) - 1;
  d.indirecttext.length = 12;

  icon_createforgetful(VTI_FOREC, 0, Y - 44, VTI_AWWIDTH, Y, 0x17000111, & d);
  d.indirecttext.text = box;
  d.indirecttext.valid = indval;
  d.indirecttext.length = 12;

  Y -= 48;
  i = icon_create(VTI_FOREC, 8 + 40, Y - 44, VTI_AWWIDTH - 56 - 44, Y, 0x1700313d, & d);
  memcpy(help[_ih], hlp, 4);
  _h[_ih ++ ] = icon_handle(i);

  Y -= 52;
  return i;
}

icon VTi_CreateLongWriteable(char * txt, char * box, int len, int type, char * hlp)
{
  icon_data d;
  icon i;

  memcpy(d.text, txt, 12);
  icon_createforgetful(VTI_FOREC, 0, Y - 44, 192, Y, 0x17000011, & d);
  Y -= 48;
  d.indirecttext.text = box;
  d.indirecttext.length = len;

  switch(type) {
  case VTI_ALL :
    d.indirecttext.valid = valid;
    break;
  case VTI_DIGITS :
    d.indirecttext.valid = numval;
    break;
  case VTI_ALPHA :
    d.indirecttext.valid = alpval;
    break;
  case VTI_ALNUM :
    d.indirecttext.valid = alnumval;
    break;
  case VTI_HEX :
    d.indirecttext.valid = hexval;
    break;
  case VTI_GROUP :
    d.indirecttext.valid = ngroupval;
    break;
  case VTI_IP :
    d.indirecttext.valid = ipval;
    break;
  }

  i = icon_create(VTI_FOREC, 9, Y - 44, VTI_AWWIDTH - 8, Y, 0x0700f13d, & d);
  icon_addhandler(i, ICON_KEYHANDLER, ick_h);
  memcpy(help[_ih], hlp, 4);
  _h[_ih ++ ] = icon_handle(i);
  /* if(!first_writeable) icon_setcaret(first_writeable = i, 0, 0, 0); */

  Y -= 52;
  /* _m[ic] = 0;
     _i[ic][1] = 0;
     _i[ic ++ ][0] = i; */

  return i;
}

icon VTi_CreateWriteable(char * txt, char * box, int len, int type, char * hlp)
{
  icon_data d;
  icon i;

  memcpy(d.text, txt, 12);
  icon_createforgetful(VTI_FOREC, 0, Y - 44, 192, Y, 0x17000211, & d);
  d.indirecttext.text = box;
  d.indirecttext.length = len;

  switch(type) {
  case VTI_ALL :
    d.indirecttext.valid = valid;
    break;
  case VTI_DIGITS :
    d.indirecttext.valid = numval;
    break;
  case VTI_ALPHA :
    d.indirecttext.valid = alpval;
    break;
  case VTI_ALNUM :
    d.indirecttext.valid = alnumval;
    break;
  case VTI_HEX :
    d.indirecttext.valid = hexval;
    break;
  case VTI_GROUP :
    d.indirecttext.valid = ngroupval;
    break;
  case VTI_IP :
    d.indirecttext.valid = ipval;
    break;
  }

  i = icon_create(VTI_FOREC, 192, Y - 44, VTI_AWWIDTH - 96, Y, 0x0700f13d, & d);
  icon_addhandler(i, ICON_KEYHANDLER, ick_h);
  memcpy(help[_ih], hlp, 4);
  _h[_ih ++ ] = icon_handle(i);
  /* if(!first_writeable) icon_setcaret(first_writeable = i, 0, 0, 0); */

  Y -= 52;
  /* _m[ic] = 0;
     _i[ic][1] = 0;
     _i[ic ++ ][0] = i; */

  return i;
}

icon VTi_CreateExample(char * txt, char * hlp)
{
  icon i;
  icon_data d;

  memcpy(d.text, txt, 12);
  icon_createforgetful(VTI_FOREC, 0, Y - 72, 192, Y, 0x17000211, & d);
  memcpy(d.text, "AaBbCcDd1234", 12);
  i = icon_create(VTI_FOREC, 192, Y - 72, VTI_AWWIDTH - 8, Y, 0x0700003d, & d);
  memcpy(help[_ih], hlp, 4);
  _h[_ih ++ ] = icon_handle(i);

  Y -= 80;
  return i;
}

int VTi_Position(int size)
{
  Y -= size;
  return Y + 8;
}
